#include "protocol.h"
#include "parser.h"
#include "kioentries.h"
#include <kdebug.h>
#include <sys/stat.h>

using namespace VFolder;
using namespace KIO;

VFolderProtocol::VFolderProtocol(QCString const &pool, QCString const &app):SlaveBase("vfolder", pool, app),root(0),desktop(0),currentXml(QString::null)
{
	kdDebug(7122) << "VFolderProtocol constructor" << endl;
}

void VFolderProtocol::readXmlFile(QString filename)
{
	kdDebug(7122) << "Parsing vfolder XML file " << filename << endl;
	if(!filename.contains('/'))
		filename="/etc/X11/desktop-menus/" + filename;
	QFile xmlFile(filename);
	if(root)
		delete root;
	root=new VFolderEntry();
	if(desktop)
		delete desktop;
#if 0
	desktop=new DesktopFiles();
#else
	// The SPECIFICATION mandrates THIS, but what ****ed gnome does
	// doesn't match it...
	desktop=new DesktopFiles("/usr/share/applications", false);
	if(getenv("DESKTOP_FILE_PATH"))
		desktop->addDirectories(getenv("DESKTOP_FILE_PATH"), false);
#endif

	VFolderParser parser(desktop, root);
	QXmlInputSource source(&xmlFile);
	QXmlSimpleReader reader;
	XmlErrorHandler errors;
	reader.setContentHandler(&parser);
	reader.setErrorHandler(&errors);
	reader.parse(source);
	currentXml=basename(QFile::encodeName(filename));
	kdDebug(7122) << "XML parser finished" << endl;
}

VFolderProtocol::~VFolderProtocol()
{
	kdDebug(7122) << "VFolderProtocol destructor" << endl;
	delete desktop;
	kdDebug(7122) << "VFolderProtocol destructor finished" << endl;
}

void VFolderProtocol::get(KURL const &url)
{
	kdDebug(7122) << "VFolder::get(" << url.path() << ")" << endl;
	error(ERR_DOES_NOT_EXIST, url.path());
	return;
}

void VFolderProtocol::stat(KURL const &url)
{
	kdDebug(7122) << "VFolder::stat(" << url.path() << ")" << endl;
	if(url.path().right(1).stripWhiteSpace()=="/")
		statEntry(DirEntry(url.path()));
	else
		statEntry(FileEntry(url.path()));
	finished();
}

void VFolderProtocol::listDir(KURL const &url)
{
	kdDebug(7122) << "VFolder::listDir(" << url.path() << ")" << endl;
	QString path(url.path());

	if(!path.contains('/')) {
		// Theoretically this can't happen, but...
		error(ERR_DOES_NOT_EXIST, path);
		return;
	}

	if(path.right(1)=="/")
		path=path.left(path.length()-1);
	if(path.left(1)=="/")
		path=path.mid(1);

	QString first;
	if(path.contains('/')) {
		first=path.left(path.find('/'));
	} else
		first=path;

	if(first.right(5)==".menu") {
		if(currentXml != first)
			readXmlFile(first);
		if(path.contains('/'))
			path=path.mid(path.find('/'));
		else
			path=QString::null;
	} else if(currentXml != "applications.menu") {
		readXmlFile();
	}

	if(path.left(1)=="/")
		path=path.mid(1);

	kdDebug(7122) << "Parsing " << first << " -/- " << path << endl;

	VFolderEntries menu=root->children();
	VFolderEntry *QuerySource=root;

	// Locate the correct subdirectory...
	while(!path.isEmpty()) {
		QString base;
		if(path.contains('/')) {
			base=path.left(path.find('/'));
			path=path.mid(path.find('/')+1);
		} else {
			base=path;
			path=QString::null;
		}

		bool found=false;
		for(VFolderEntries::const_iterator it=menu.begin(); it!=menu.end(); it++)
			if((*it)->name() == base) {
				QuerySource=*it;
				menu=(*it)->children();
				found=true;
				break;
			}

		if(!found) {
			error(ERR_DOES_NOT_EXIST, url.path());
			return;
		}
	}

	// remove all empty folders
	checkFolder(QuerySource);
	menu = QuerySource->children();
	
	// Add submenus
	for(VFolderEntries::const_iterator it=menu.begin(); it!=menu.end(); it++) {
		if((*it)->type() == Dir)
			listEntry(DirEntry((*it)->name(), (*it)->desktop(), "vfolder:" + url.path() + "/"), false);
		else // This shouldn't happen, but...
			listEntry(FileEntry((*it)->name()), false);
	}

	// Check desktop files for category matches...
	for(DesktopFiles::iterator it=desktop->begin(); it!=desktop->end(); it++) {
		if(QuerySource->match(*(*it)) && ! (*it)->fileName().contains(".directory")) {
			struct stat s;
			::stat(QFile::encodeName((*it)->fileName()), &s);
			listEntry(FileEntry((*it)->readName(), s.st_mode,
						 s.st_size, (*it)->fileName(), "application/x-desktop"), false);
		}
	}

	UDSEntry e;
	listEntry(e, true);
	finished();
}

bool VFolderProtocol::checkFolder(VFolderEntry *folder) {
	if (folder->type() == File)
		return true;

	VFolderEntries subfolders = folder->children();
	bool has_items = false;
  
	for(VFolderEntries::const_iterator it=subfolders.begin(); it!=subfolders.end(); it++)
		if (checkFolder(*it)) {
			has_items = true;
			break;
		}

	for(DesktopFiles::iterator it=desktop->begin(); it!=desktop->end(); it++)
		if (folder->match(*(*it))) {
			has_items = true;
			break;
		}

	if (folder->dontShowIfEmpty() && has_items == false) {
		folder->parent()->children().remove(folder);
		delete folder;
	}

   return false;
}
